﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using STSdb.Data;
using STSdb.Plan;
using System.Diagnostics;
using System.IO;
using STSdb.Files;
using System.Windows.Forms;

namespace GettingStarted
{
    public class GettingStarted
    {
        public StorageEngine Repository { get; private set; }
        public string FileName { get; private set; }
        public ulong RecordCount { get; set; }

        public GettingStarted(string fileName)
        {
            FileName = fileName;
            RecordCount = 1000;
        }

        /// <summary>
        /// Open database
        /// </summary>
        public void OpenRepository()
        {
            //Open database
            Repository = StorageEngine.FromFile(FileName);
        }

        /// <summary>
        /// Add records
        /// </summary>
        public void Append(Locator locator)
        {
            //create or open table
            XTable<ulong, Tick> table = Repository.Scheme.CreateOrOpenXTable<ulong, Tick>(locator);

            //table.RecordPersist = new TickBinaryPersist(); //if you want some custom storage implementation.
            //table.RecordPersist = new DeflateTickBinaryPersist(); //demonstrates deflate compression over custom storage solution

            Repository.Scheme.Commit();

            ulong id = table.Count == 0 ? 0 : table.LastRow.Key + 1;
            foreach (Tick tick in Tick.GenerateRandom(RecordCount))
            {
                //insert record
                table[id] = tick;
                id++;
            }
            table.Commit();

            //close table
            table.Close();
        }

        /// <summary>
        /// Sort by Timestamp
        /// </summary>
        public Locator SortByTimestamp(Locator locator)
        {
            //open source table
            XTable<ulong, Tick> srcTable = Repository.Scheme.OpenXTable<ulong, Tick>(locator);

            //create a stress buffer for sorting operation
            StressBuffer<XKey<DateTime, ulong>, Tick> buffer = new StressBuffer<XKey<DateTime, ulong>, Tick>(Repository);
            
            //add data to the stress buffer
            ulong id = 0;
            foreach (var row in srcTable)
                buffer.Add(new XKey<DateTime,ulong>(row.Record.Timestamp, id++), row.Record);

            //close source table
            srcTable.Close();

            //recreate destination table
            Locator dstLocator = locator.Append("Sort", "Timestamp");
            Repository.Scheme.Delete(dstLocator);
            XTable<XKey<DateTime, ulong>, Tick> dstTable = Repository.Scheme.CreateOrOpenXTable<XKey<DateTime, ulong>, Tick>(dstLocator);
            Repository.Scheme.Commit();

            //merge data
            foreach (var row in buffer)
                dstTable[row.Key] = row.Record;
            dstTable.Commit();

            //close table
            dstTable.Close();

            return dstLocator;
        }

        /// <summary>
        /// Sort by Bid
        /// </summary>
        public Locator SortByBid(Locator locator)
        {
            //open source table
            XTable<ulong, Tick> srcTable = Repository.Scheme.OpenXTable<ulong, Tick>(locator);

            //create a stress buffer for sorting operation
            StressBuffer<XKey<double, ulong>, Tick> buffer = new StressBuffer<XKey<double, ulong>, Tick>(Repository);

            //add data to the stress buffer
            ulong id = 0;
            foreach (var row in srcTable)
                buffer.Add(new XKey<double, ulong>(row.Record.Bid, id++), row.Record);

            //close source table
            srcTable.Close();

            //recreate destination table
            Locator dstLocator = locator.Append("Sort", "Bid");
            Repository.Scheme.Delete(dstLocator);
            XTable<XKey<double, ulong>, Tick> dstTable = Repository.Scheme.CreateOrOpenXTable<XKey<double, ulong>, Tick>(dstLocator);
            Repository.Scheme.Commit();

            //sort data
            foreach (var row in buffer)
                dstTable[row.Key] = row.Record;
            dstTable.Commit();

            //close table
            dstTable.Close();

            return dstLocator;
        }

        /// <summary>
        /// Sort By Exchange Then By Timestamp
        /// </summary>
        public Locator SortByExchangeThenByTimestamp(Locator locator)
        {
            //open source table
            XTable<ulong, Tick> srcTable = Repository.Scheme.OpenXTable<ulong, Tick>(locator);

            //creates a stress buffer for sorting operation
            StressBuffer<XKey<string, DateTime, ulong>, Tick> buffer = new StressBuffer<XKey<string, DateTime, ulong>, Tick>(Repository);

            //add data to the stress buffer
            ulong id = 0;
            foreach (var row in srcTable)
                buffer.Add(new XKey<string, DateTime, ulong>(row.Record.Exchange, row.Record.Timestamp, id++), row.Record);

            //close source table
            srcTable.Close();

            //recreate destination table
            Locator dstLocator = locator.Append("Sort", "Exchange&Timestamp");
            Repository.Scheme.Delete(dstLocator);
            XTable<XKey<string, DateTime, ulong>, Tick> dstTable = Repository.Scheme.CreateOrOpenXTable<XKey<string, DateTime, ulong>, Tick>(dstLocator);
            Repository.Scheme.Commit();

            //merge data
            foreach (var row in buffer)
                dstTable[row.Key] = row.Record;
            dstTable.Commit();

            //close table
            dstTable.Close();

            return dstLocator;
        }

        /// <summary>
        /// Partition By Exchange
        /// </summary>
        public Locator PartitionByExchangeAndSortByTimestamp(Locator locator)
        {
            //open source table
            XTable<ulong, Tick> srcTable = Repository.Scheme.OpenXTable<ulong, Tick>(locator);

            //creates a stress buffer for partitioning & sorting operations
            StressBuffer<XKey<string, DateTime, ulong>, Tick> buffer = new StressBuffer<XKey<string, DateTime, ulong>, Tick>(Repository);

            //add data to the stress buffer
            ulong id = 0;
            foreach (var row in srcTable)
                buffer.Add(new XKey<string, DateTime, ulong>(row.Record.Exchange, row.Record.Timestamp, id++), row.Record);

            //close source table
            srcTable.Close();

            string exchange = null;
            XTable<XKey<DateTime, ulong>, Tick> dstTable = null;

            foreach (var row in buffer)
            {
                if (exchange == null || exchange != row.Key.SubKey0)
                {
                    exchange = row.Key.SubKey0;

                    Locator dstTableName = locator.Append("Partition", "Exchange", exchange);
                    Repository.Scheme.Delete(dstTableName);
                    if (dstTable != null)
                    {
                        dstTable.Commit();
                        dstTable.Close();
                    }
                    dstTable = Repository.Scheme.CreateOrOpenXTable<XKey<DateTime, ulong>, Tick>(dstTableName);
                }

                dstTable[new XKey<DateTime, ulong>(row.Key.SubKey1, row.Key.SubKey2)] = row.Record;
            }

            if (dstTable != null)
            {
                dstTable.Commit();
                dstTable.Close();
                Repository.Scheme.Commit();
            }

            return locator.Append("Partition", "Exchange"); //root of destination tables
        }

        /// <summary>
        /// Partition By Exchange And Symbol
        /// </summary>
        public Locator PartitionByExchangeAndSymbolAndSortByTimestamp(Locator locator)
        {
            //open source table
            XTable<ulong, Tick> srcTable = Repository.Scheme.OpenXTable<ulong, Tick>(locator);

            //creates a stress buffer for partitioning & sorting operation
            StressBuffer<XKey<string, string, DateTime, ulong>, Tick> buffer = new StressBuffer<XKey<string, string, DateTime, ulong>, Tick>(Repository);

            //add data to the stress buffer
            ulong id = 0;
            foreach (var row in srcTable)
                buffer.Add(new XKey<string, string, DateTime, ulong>(row.Record.Exchange, row.Record.Symbol, row.Record.Timestamp, id++), row.Record);

            //close source table
            srcTable.Close();

            string exchange = null;
            string symbol = null;
            XTable<XKey<DateTime, ulong>, Tick> dstTable = null;

            foreach (var row in buffer)
            {
                if (exchange == null || symbol == null ||
                    exchange != row.Key.SubKey0 || symbol != row.Key.SubKey1)
                {
                    exchange = row.Key.SubKey0;
                    symbol = row.Key.SubKey1;

                    Locator dstTableName = locator.Append("Partition", "Exchange&Symbol", exchange, symbol);
                    Repository.Scheme.Delete(dstTableName);
                    if (dstTable != null)
                    {
                        dstTable.Commit();
                        dstTable.Close();
                    }
                    dstTable = Repository.Scheme.CreateOrOpenXTable<XKey<DateTime, ulong>, Tick>(dstTableName);
                }

                dstTable[new XKey<DateTime, ulong>(row.Key.SubKey2, row.Key.SubKey3)] = row.Record;
            }

            if (dstTable != null)
            {
                dstTable.Commit();
                dstTable.Close();
                Repository.Scheme.Commit();
            }

            return locator.Append("Partition", "Exchange&Symbol"); //root of destination tables
        }

        /// <summary>
        /// Sort By Exchange Desc Then By Timestamp 
        /// </summary>
        public Locator SortByExchangeDescThenByTimestamp(Locator locator)
        {
            //recreate destination table
            Locator dstLocator = locator.Append("Sort", "Exchange Descending & Timestamp");
            Repository.Scheme.Delete(dstLocator);
            XTable<XKey<string, DateTime, ulong>, Tick> dstTable = Repository.Scheme.CreateOrOpenXTable<XKey<string, DateTime, ulong>, Tick>(dstLocator);
            dstTable.KeyMap = new XKeyMap<string, DateTime, ulong>(new StringKeyMap(64, SortOrder.Descending), 0, null, 0, null, 0);
            Repository.Scheme.Commit();

            //open source table
            XTable<ulong, Tick> srcTable = Repository.Scheme.OpenXTable<ulong, Tick>(locator);

            //creates a stress buffer for sort operation
            StressBuffer<XKey<string, DateTime, ulong>, Tick> buffer = new StressBuffer<XKey<string, DateTime, ulong>, Tick>(Repository, dstTable.KeyMap, null);

            //add data to the stress buffer
            ulong id = 0;
            foreach (var row in srcTable)
                buffer.Add(new XKey<string, DateTime, ulong>(row.Record.Exchange, row.Record.Timestamp, id++), row.Record);

            //close source table
            srcTable.Close();

            //merge data
            foreach (var row in buffer)
                dstTable[row.Key] = row.Record;
            dstTable.Commit();

            //close tables
            dstTable.Close();

            return dstLocator;
        }

        /// <summary>
        /// Read records
        /// </summary>
        public Locator Read(Locator locator)
        {
            //open table
            XTable<ulong, Tick> table = Repository.Scheme.OpenXTable<ulong, Tick>(locator);
            if (table == null)
                return locator;

            //read in ascending order
            foreach (Row<ulong, Tick> row in table.Forward()) //table.Backward()
            {
                //do something
            }

            //close table
            table.Close();

            return locator;
        }

        /// <summary>
        /// Index by Timestamp
        /// </summary>
        private Locator IndexByTimestamp(Locator locator)
        {
            //open source table
            XTable<ulong, Tick> srcTable = Repository.Scheme.OpenXTable<ulong, Tick>(locator);

            //create a stress buffer for sorting operation
            StressBuffer<XKey<DateTime, ulong>, ulong> buffer = new StressBuffer<XKey<DateTime, ulong>, ulong>(Repository);

            //add data to the stress buffer
            ulong id = 0;
            foreach (var row in srcTable)
            {
                buffer.Add(new XKey<DateTime, ulong>(row.Record.Timestamp, id), row.Key);
                id++;
            }

            //close source table
            srcTable.Close();

            //recreate index table
            Locator dstLocator = locator.Append("Index", "Timestamp");
            Repository.Scheme.Delete(new Locator(dstLocator));
            XTable<XKey<DateTime, ulong>, ulong> dstTable = Repository.Scheme.CreateOrOpenXTable<XKey<DateTime, ulong>, ulong>(dstLocator);
            Repository.Scheme.Commit();

            //merge data
            foreach (var row in buffer)
                dstTable[row.Key] = row.Record;
            dstTable.Commit();

            //close table
            dstTable.Close();

            return dstLocator;
        }

        /// <summary>
        /// Read through index
        /// </summary>
        private Locator ReadThroughIndex(Locator locator)
        {
            //open table
            XTable<ulong, Tick> table = Repository.Scheme.OpenXTable<ulong, Tick>(locator);

            //open index table
            Locator indexLocator = locator.Append("Index", "Timestamp");
            XTable<XKey<DateTime, ulong>, ulong> indexTable = Repository.Scheme.OpenXTable<XKey<DateTime, ulong>, ulong>(indexLocator);

            //read trough indexTable
            foreach (var indexRow in indexTable.Forward())
            {
                Tick record = table[indexRow.Record];
            }

            //close tables
            table.Close();
            indexTable.Close();

            return locator;
        }


        /// <summary>
        /// create and read through index
        /// </summary>
        public Locator CreateAndReadThroughIndex(Locator locator)
        {
            Locator indexLocator = IndexByTimestamp(locator);
            ReadThroughIndex(locator);

            return indexLocator;
        }

        /// <summary>
        /// Update Ask
        /// </summary>
        public Locator UpdateAsk(Locator locator)
        {
            //open source table
            XTable<ulong, Tick> table = Repository.Scheme.CreateOrOpenXTable<ulong, Tick>(locator);

            //take a snapshot of the table
            XTable<ulong, Tick> snapshot = table.Snapshot();

            //iterate data through snapshot and update through table
            foreach (var row in snapshot)
            {
                //update record (set fixed spread)
                row.Record.Ask = row.Record.Bid + 4 * 0.0001;

                table[row.Key] = row.Record;
            }
            table.Commit();

            //close
            table.Close();

            //if you want to keep and develop table snapshot as a separate branch, then register it in Repository Scheme.
            //Repository.Scheme.Register(locator.Append("original"), snapshot);

            return locator;
        }

        public Locator InsertBlobData(Locator locator)
        {
            //to import large data objects we will use repository file system - the underlying database layer.
            
            Locator dstLocator = locator.Append("Blob");
            XTable<ulong, BlobPointer> table = Repository.Scheme.CreateOrOpenXTable<ulong, BlobPointer>(dstLocator);
            ulong id = table.Count;
            
            //we will import files from directory..
            string directory = @"C:\Program Files";
            FolderBrowserDialog dialog = new FolderBrowserDialog();
            dialog.SelectedPath = directory;
            if (dialog.ShowDialog() == DialogResult.OK)
                directory = dialog.SelectedPath;
            else
                return dstLocator;

            //..in single repository file
            RawFile rawFile = Repository.FileSystem.CreateFile();
            BinaryFile binaryFile = new BinaryFile(rawFile);

            try
            {
                foreach (var fileName in Directory.GetFiles(directory, "*.*", SearchOption.AllDirectories))
                {
                    using (FileStream fs = new FileStream(fileName, FileMode.Open, FileAccess.Read))
                    {
                        //create blob pointer
                        BlobPointer pointer = new BlobPointer();
                        pointer.Handle = rawFile.Handle;
                        pointer.Offset = binaryFile.Position;
                        pointer.Length = (ulong)fs.Length;
                        pointer.FileName = fileName;

                        //import file content
                        binaryFile.Write(fs);

                        //add blob pointer to table
                        table[id++] = pointer;
                    }
                }

                table.Commit();
            }
            finally
            {
                binaryFile.Close();
                Repository.Scheme.Commit();
            }

            return dstLocator;
        }

        public class BlobPointer
        {
            public ulong Handle { get; set; }
            public ulong Offset { get; set; }
            public ulong Length { get; set; }
            public string FileName { get; set; }

            public BlobPointer()
            {
            }
        }

        /// <summary>
        /// Close database
        /// </summary>
        public void CloseRepository()
        {
            //close database
            Repository.Dispose();
        }

        public ulong GetRecordCount(Locator locator)
        {
            var table = Repository.Scheme.OpenXTable(locator);
            if (table == null)
                return 0;

            return table.Count;
        }
    }
}