﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ComponentModel;
using General.Collections;
using STSdb.Persist;
using System.IO;
using System.IO.Compression;

namespace GettingStarted
{
    public class Tick
    {
        public string Symbol { get; set; }
        public DateTime Timestamp { get; set; }
        public double Bid { get; set; }
        public double Ask { get; set; }
        public long Volume { get; set; }
        public string Exchange { get; set; }

        public Tick()
        {
        }

        /// <summary>
        /// Generate random ticks
        /// </summary>
        public static IEnumerable<Tick> GenerateRandom(ulong count)
        {
            Random random = new Random();

            //Some symbols
            string[] symbols = new string[] { "EURUSD", "GBPCHF", "EURGBP", "JPYUSD", "GBPCAD" };
            string[] exchanges = new string[] { "NYSE", "TSE", "NASDAQ", "Euronext", "LSE", "SSE", "ASE", "SE", "NSEI" };
            double[] prices = new double[symbols.Length];
            double[] pipsizes = new double[symbols.Length];
            int[] digits = new int[symbols.Length];

            //Initialize startup data
            for (int i = 0; i < symbols.Length; i++)
            {
                prices[i] = 1.5;
                pipsizes[i] = 0.0001;
                digits[i] = 4;
            }

            DateTime timestamp = DateTime.Now;

            //Generate ticks
            for (ulong i = 0; i < count; i++)
            {
                int id = random.Next(symbols.Length);

                //generate random movement
                int pips = random.Next(0, 10);
                int direction = random.Next() % 2 == 0 ? 1 : -1;
                int spread = random.Next(2, 30);
                int seconds = random.Next(1, 30);

                //generate values
                string symbol = symbols[id];

                timestamp = timestamp.AddSeconds(seconds);

                double bid = prices[id];
                bid += direction * pips * pipsizes[id];
                bid = Math.Round(bid, digits[id]);

                double ask = bid + spread * pipsizes[id];
                ask = Math.Round(ask, digits[id]);

                //create tick
                Tick tick = new Tick();
                tick.Symbol = symbol;
                tick.Timestamp = timestamp;
                tick.Bid = bid;
                tick.Ask = ask;
                tick.Volume = random.Next(0, 10000);
                tick.Exchange = exchanges[random.Next(exchanges.Length)];

                yield return tick;
            }
        }

        public override string ToString()
        {
            return string.Format("{0} {1}/{2}({3}) at {4} on {5}", Symbol, Bid, Ask, Volume, Timestamp, Exchange);
        }
    }

    /// <summary>
    /// if you want some custom storage implementation.
    /// </summary>
    [Serializable]
    public class TickBinaryPersist : IBinaryPersist<IIndexer<Tick>>
    {
        #region IBinaryPersist<IIndexer<Tick>> Members

        public void Load(BinaryReader reader, ref IIndexer<Tick> data)
        {
            for (int i = 0; i < data.Count; i++)
            {
                Tick tick = new Tick();

                tick.Symbol = reader.ReadString();
                tick.Timestamp = new DateTime(reader.ReadInt64());
                tick.Bid = reader.ReadDouble();
                tick.Ask = reader.ReadDouble();
                tick.Exchange = reader.ReadString();
                tick.Volume = reader.ReadInt64();

                data[i] = tick;
            }
        }

        public void Store(BinaryWriter writer, IIndexer<Tick> data)
        {
            for (int i = 0; i < data.Count; i++)
            {
                Tick tick = data[i];

                writer.Write(tick.Symbol);
                writer.Write(tick.Timestamp.Ticks);
                writer.Write(tick.Bid);
                writer.Write(tick.Ask);
                writer.Write(tick.Exchange);
                writer.Write(tick.Volume);
            }
        }

        #endregion
    }

    /// <summary>
    /// demonstrates deflate compression over custom storage solution
    /// </summary>
    [Serializable]
    public class DeflateTickBinaryPersist : IBinaryPersist<IIndexer<Tick>>
    {
        public readonly IBinaryPersist<IIndexer<Tick>> Persist;

        public DeflateTickBinaryPersist(IBinaryPersist<IIndexer<Tick>> persist)
        {
            this.Persist = persist;
        }

        public DeflateTickBinaryPersist()
            :this(new TickBinaryPersist())
        {
        }
        
        #region IBinaryPersist<IIndexer<Tick>> Members

        public void Load(BinaryReader reader, ref IIndexer<Tick> data)
        {
            int length = reader.ReadInt32();
            using (MemoryStream memoryStream = new MemoryStream(reader.ReadBytes(length)))
            {
                using (DeflateStream deflateStream = new DeflateStream(memoryStream, CompressionMode.Decompress, true))
                    Persist.Load(new BinaryReader(deflateStream), ref data);
            }
        }

        public void Store(BinaryWriter writer, IIndexer<Tick> data)
        {
            using (MemoryStream memoryStream = new MemoryStream())
            {
                using (DeflateStream deflateStream = new DeflateStream(memoryStream, CompressionMode.Compress, true))
                    Persist.Store(new BinaryWriter(deflateStream), data);

                writer.Write((int)memoryStream.Length);
                writer.Write(memoryStream.GetBuffer(), 0, (int)memoryStream.Length);
            }            
        }

        #endregion
    }
}
